#define DETERMINISTIC
#include "GC/gc.h"
#include "header.h"

/* For exception handling, a global storing the state information
 * needed to longjmp back to the dynamically most recent protect
 * statement.  Also needed is a global for returning the exception
 * value.
*/

jmp_buf last_protect;
void *exception;

static INT ob_count;
extern BOOL deterministic;

void setbuffer(FILE *fp, int sz)                                             /* -- NLP */
{                                                                            /* -- NLP */
   setvbuf(fp, NULL, _IOFBF, sz);                                            /* -- NLP */
}                                                                            /* -- NLP */

int execve(char *path, char *argv[], char *envp[])                           /* -- NLP */
{                                                                            /* -- NLP */
   return _execve(path, argv, envp);                                         /* -- NLP */
}                                                                            /* -- NLP */

void mkdirxx(char *p)                                                        /* -- NLP */
{                                                                            /* -- NLP */
   _mkdir(p);                                                                /* -- NLP */
}                                                                            /* -- NLP */

double rint(double x)                                                        /* -- NLP */
{                                                                            /* -- NLP */
        double c,f;                                                          /* -- NLP */
        c = ceil(x);                                                         /* -- NLP */
        f = floor(x);                                                        /* -- NLP */
        if (c == f) return x;                                                /* -- NLP */
        if ((c - x) < (x - f)) return c;                                     /* -- NLP */
        return f;                                                            /* -- NLP */
}                                                                            /* -- NLP */


void *sbi_alloc(size_t size, INT tag) {
    void *res;
    /* printf("sbi_alloc %d %d\n",size,tag); */
    res=GC_malloc(size);
    if (res==NULL) {
	fprintf(stderr,"Fatal runtime error: unable to allocate more memory");
/*      abort();                                                                -- NLP */
        exit(16);                                                         /*    -- NLP */
    }
    ((OB)res)->header.tag=tag;
    if (deterministic) ((OB)res)->header.id=ob_count++;
    return res;
}

void *sbi_arr_alloc(size_t size1, INT tag, size_t size2, INT n) {
    void *res;
    /* printf("sbi_arr_alloc %d %d %d %d\n",size1,tag,size2,n); */
    res=GC_malloc(size1+(n-1)*size2);
    if (res==NULL) {
	fprintf(stderr,"Fatal runtime error: unable to allocate more memory");
/*      abort();                                                                -- NLP */
        exit(16);                                                         /*    -- NLP */
    }
    ((OB)res)->header.tag=tag;
    if (deterministic) ((OB)res)->header.id=ob_count++;
    return res;
}

void *sbi_alloc_atomic(size_t size, INT tag) {
     void *res;
     /* printf("sbi_alloc_atomic %d %d\n",size,tag); */
/*   res=GC_malloc_atomic(size);                                                -- NLP */
     res=GC_malloc(size);                                                 /*    -- NLP */
     /* we must clear it manually */
/*   memset(res,0,size);                                                        -- NLP */
     if (res==NULL) {
       fprintf(stderr,"Fatal runtime error: unable to allocate more memory");
/*     abort();                                                                 -- NLP */
       exit(16);                                                          /*    -- NLP */
     }
     ((OB)res)->header.tag=tag;
     if (deterministic) ((OB)res)->header.id=ob_count++;
     return res;
}
 
void *sbi_arr_alloc_atomic(size_t size1, INT tag, size_t size2, INT n) {
     void *res;
     /* printf("sbi_arr_alloc %d %d %d %d\n",size1,tag,size2,n); */
     size_t size = size1+(n-1)*size2;
/*   res=GC_malloc_atomic(size);                                                -- NLP */
     res=GC_malloc(size);                                                 /*    -- NLP */
/*   memset(res,0,size);                                                        -- NLP */
     if (res==NULL) {
       fprintf(stderr,"Fatal runtime error: unable to allocate more memory");
/*     abort();                                                                 -- NLP */
       exit(16);                                                          /*    -- NLP */
     }
     ((OB)res)->header.tag=tag;
     if (deterministic) ((OB)res)->header.id=ob_count++;
     return res;
}
 

/* This catches void references */

void sbi_segfault_handler()
{
    fprintf(stderr,"Fatal error: Attribute or array access of void.\n");
/*  abort();                                                                    -- NLP */
    exit(16);                                                             /*    -- NLP */
}

/* Set (or clear) a bit.  It's a fatal error for bit to be less than
 * zero or greater than or equal to maxbits.  This implementation
 * is probably endian dependent.  If anyone knows a more portable
 * way to do this, let me know.
*/

void sbi_aset(void *buf, INT bit, BOOL val, INT maxbits)
{
    INT i,j,mask,*p;
    if (bit<0||bit>=maxbits) {
	fprintf(stderr,"Fatal error: aset index out of range\n");
/*      abort();                                                                -- NLP */
        exit(16);                                                         /*    -- NLP */
	}
    i = bit/(sizeof(INT)*8);
    j = bit%(sizeof(INT)*8);
    mask = ~(1<<j);
    p = ((INT*) buf)+i;
    *p = ((*p)&mask)|(val<<j);
}

BOOL sbi_aget(void *buf, INT bit, INT maxbits)
{
    INT i,j,*p;
    if (bit<0||bit>=maxbits) {
	fprintf(stderr,"Fatal error: aget index out of range\n");
/*      abort();                                                                -- NLP */
        exit(16);                                                         /*    -- NLP */
	}
    i = bit/(sizeof(INT)*8);
    j = bit%(sizeof(INT)*8);
    p = ((INT*) buf)+i;
    return (BOOL) (((*p)&(1<<j))!=0);
}
